/*
 *	%Z% %M% %I% %D% %Q%
 *
 *	Copyright (C) The Santa Cruz Operation, 1985, 1986.
 *	This Module contains Proprietary Information of
 *	The Santa Cruz Operation and should be treated as Confidential.
 *
 */

/*
 *    AT&T  real-time clock driver
 * 
 *    This driver converts a date given in asktime(C) format
 *    to a format close to that used by the AT&T 6300 real-time clock.
 *    Neither the 6300 day-of-week nor the 6300 leap-year register is used.
 *    Since the AT&T clock/calender keeps up to eight years, and XENIX has
 *    a longer view of the current year, the driver simply returns its
 *    idea of the year, from a base of 1984.  This driver will break
 *    in 1992. 
 */

#include "../h/param.h"
#include "../h/dir.h"
#include "../h/user.h"
#include "../h/systm.h"

#define TEST_PORT     	0x70   /*   (W)    0=not test mode, 1=test mode */
#define TENTH_SECONDS 	0x71   /*   (R)    0-9  */
#define SECONDS       	0x72   /*   (R)    0-9  */
#define TENS_SECONDS  	0x73   /*   (R)    0-5  */
#define MINUTES       	0x74   /*   (R/W)  0-9  */
#define TENS_MINUTES  	0x75   /*   (R/W)  0-9  */
#define HOURS         	0x76   /*   (R/W)  0-9  */
#define TENS_HOURS    	0x77   /*   (R/W)  0-2  */
#define DAYS          	0x78   /*   (R/W)  0-9  */
#define TENS_DAYS     	0x79   /*   (R/W)  0-3  */
#define DAY_OF_WEEK   	0x7A   /*   (R/W)  1-7  */
#define MONTHS        	0x7B   /*   (R/W)  0-9  */
#define TENS_MONTHS   	0x7C   /*   (R/W)  0-1  */
#define LEAP_YEAR     	0x7D   /*   (W)    0-3, years til leap year + 1 */
#define STOP_START     	0x7E   /*   (W)    0=stop, 0xFF=start */
#define YEARS         	0x7F   /*   (R/W)  year mod 8 */

#define RTCPRI          (PZERO + 8)
#define STOP            0x00
#define START           0xFF
#define YEARMASK        0x07
#define ASCII_ZERO      0x30
#define CLOCKUPDATE     0x0F
#define CHRSTRSIZ       10
#define BASEDATE        84     /*   as in 1984, because DOS is '84 based */
#define CENTURY         1900

int seconds;

/*ARGSUSED*/
rtcopen(mdev)
{
	int s;
	static int rtcinited=0;
	int rtctimeout();

	if ( !rtcinited ) {
		seconds = (rtcreadreg(SECONDS) << 4) + 
			   rtcreadreg(TENTH_SECONDS);
		s = spl7();
		timeout(rtctimeout,0,Hz/4);
		sleep((caddr_t)&seconds,RTCPRI);
		splx(s);
		if (( rtcreadreg(SECONDS) << 4 ) + 
		      rtcreadreg(TENTH_SECONDS) == seconds ) {
			   u.u_error = ENXIO;
			   return;
		}
		rtcinited++;
	}
}

rtctimeout()
{
	wakeup((caddr_t)&seconds);
}

/*
 * rtcread -- read real time clock 
 */
/*ARGSUSED*/
rtcread(mdev)
{
	register int i;
	char datestr[CHRSTRSIZ];

	rtcgets(datestr);
	while (( i= u.u_offset ) < CHRSTRSIZ && u.u_count > 0 )
		if (passc(datestr[i]) == -1) 
			return;
	u.u_rval1 = CHRSTRSIZ;
}

/*
 * rtcwrite -- read real time clock 
 */
/*ARGSUSED*/
rtcwrite(mdev)
{
	register int i,n;
	char datestr[CHRSTRSIZ];

	if ( u.u_count < CHRSTRSIZ ) {
		u.u_error = EINVAL;
		return;
        }	
	for ( i=0 ; i < CHRSTRSIZ ; i++ ) {
		if (( n = cpass()) == -1 ) 
			return;
		datestr[i] = n;
	}
	rtcputs(datestr);
	u.u_rval1 = CHRSTRSIZ;
}

static int mid_update = 0;

/*
 * rtcgets -- get time from real time clock and convert to asktime format
 * ( except for years, which are given as year modulo 8 )
 * if an update occurred during any of the reads, read everything
 * again, as other values may have changed as well.
 */
rtcgets(datestr)
char *datestr;
{
	int year;
	int n=5;

	do { 
		mid_update  = 0;
		year = ( rtcreadreg(YEARS) & YEARMASK ) + BASEDATE;
		datestr[0]  = rtcreadreg(TENS_MONTHS)          + ASCII_ZERO;
		datestr[1]  = rtcreadreg(MONTHS)               + ASCII_ZERO;
		datestr[2]  = rtcreadreg(TENS_DAYS)            + ASCII_ZERO;
		datestr[3]  = rtcreadreg(DAYS)                 + ASCII_ZERO;
		datestr[4]  = rtcreadreg(TENS_HOURS)           + ASCII_ZERO;
		datestr[5]  = rtcreadreg(HOURS)                + ASCII_ZERO;
		datestr[6]  = rtcreadreg(TENS_MINUTES)         + ASCII_ZERO;
		datestr[7]  = rtcreadreg(MINUTES)              + ASCII_ZERO;
		datestr[8]  = year / 10                        + ASCII_ZERO;
		datestr[9]  = year % 10                        + ASCII_ZERO; 
	} while ( mid_update && n-- );
}

rtcputs(datestr)
char *datestr;
{
	int yr;

	rtcstop();
	rtcwritreg(TENS_MONTHS,datestr[0]);
	rtcwritreg(MONTHS,datestr[1]);
	rtcwritreg(TENS_DAYS,datestr[2]);
	rtcwritreg(DAYS,datestr[3]);
	rtcwritreg(TENS_HOURS,datestr[4]);
	rtcwritreg(HOURS,datestr[5]);
	rtcwritreg(TENS_MINUTES,datestr[6]);
	rtcwritreg(MINUTES,datestr[7]);
	yr = (rtc_atoi(datestr[8],datestr[9]) + CENTURY) % 8;
	rtcwritreg(LEAP_YEAR, 3 - ( yr % 4 ));
	rtcwritreg(YEARS,yr | 0x08);
	rtcreadreg(YEARS);    rtcreadreg(YEARS);    rtcreadreg(YEARS);
	rtcwritreg(DAY_OF_WEEK,rtcday_of_week());
	rtcstart();
}

/* 
 * read clock register "reg"
 * if value of "reg" is CLOCKUPDATE, the chip is currently updating
 * the time of day, so wait until it isn't updating anymore
 * and set a flag so that we know to re-read the date, as
 * other parts may have changed.
 */
rtcreadreg(reg)
{
	register int i;
	int n=0x400; /* arbitrary */

	while (( i = ( inb(reg) & CLOCKUPDATE )) == CLOCKUPDATE && n-- )
		mid_update++;
	return(i);
}

rtcwritreg(reg,val)
register reg,val;
{
	outb(reg,val - ASCII_ZERO);
}	

rtcstop()
{
	outb(TEST_PORT, 0);
	outb(STOP_START,STOP);
}

rtcstart()
{
	outb(STOP_START,START);
}

rtc_atoi(tens,ones)
{
	return(( tens - ASCII_ZERO ) * 10 + ( ones - ASCII_ZERO ));
}


/*
 * Determine day of week from information given in datestr.
 * Although this information is not used directly, it must be
 * set so that the day of the week will be correct under DOS after being set
 * under XENIX.  
 *
 * Rather than have the driver know all the fancy conversion rules needed
 * to turn a date into the number of days that have passed since the 
 * beginning of computerized time, we index into several arrays: the first 
 * contains the day of the week on which Jan 1 occurs each year; 
 * the other two tell the day of the week on which the first day of each month
 * occurs. There are separate tables for leap years and non-leap years.
 * 0 is Sunday, 6 is Saturday.
 *
 * While "year modulo 4" is not a definitive test for whether or not a
 * year is a leap year, it is valid for every year from now until
 * 2099, which is far beyond the lifetime of the driver.   
 * If a year is divisible by 100, it is not a leap
 * year unless it is also divisible by 400. 
 * Yes folks, 2000 _is_ a leap year. 
 */
                        /* Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec */
char     month_offset[] = { 0 , 3 , 3 , 6 , 1 , 4 , 6 , 2 , 5 , 0 , 3 , 5 };
char leapmonth_offset[] = { 0 , 3 , 4 , 0 , 2 , 5 , 0 , 3 , 6 , 1 , 4 , 6 };

                        /* 84  85  86  87  88  89  90  91 */
char      year_offset[] = { 0 , 2 , 3 , 4 , 5 , 0 , 1 , 2 };

rtcday_of_week()
{
        int year, month, day;

	day  = rtc_atoi(rtcreadreg(TENS_DAYS),rtcreadreg(DAYS) - 1);
	year = year_offset[rtc_atoi(0,rtcreadreg(YEARS))];
	if ( year % 4 == 0 )
		month = leapmonth_offset[rtc_atoi(rtcreadreg(TENS_MONTHS),
						  rtcreadreg(MONTHS))-1];
	else
		month = month_offset[rtc_atoi(rtcreadreg(TENS_MONTHS),
					          rtcreadreg(MONTHS))-1];
	return(((year + month + day) % 7 ) + 1);
}
